#include <stdio.h>
#include <string.h>

#include "adapter_api.h"

unsigned long BusClose(s_Adapter *p_HW)
{
	unsigned long Status = 0;

	// Close the assigned device.
	Status = Adapter_Close(p_HW);
	return(Status);
}

unsigned long BusOpen(s_Adapter *p_HW, unsigned long Bar, unsigned long Off, unsigned long long IntDisable)
{
	unsigned long Status = 0;

	// Set all of the global interrupt mask register values in the hardware handle.
	p_HW->GlobalIntMask.BAR = Bar;
	p_HW->GlobalIntMask.Offset = Off;
	p_HW->GlobalIntMask.Disable = IntDisable;
	// Open the assigned device.
	Status = Adapter_Open(p_HW);
	return(Status);
}

unsigned long DMAAllocate(s_Adapter *p_HW, unsigned long Bar, unsigned long Off, unsigned long ChannelIndex, unsigned long BurstSizeB, unsigned long BurstCount, unsigned long PageCount, unsigned long PagesPerMark)
{
	unsigned long Status = 0;
	unsigned long page_count;
	unsigned long long descriptor;
	unsigned long add_index;

	// Set the device and channel number.
	p_HW->DMA.ChannelIndex = ChannelIndex - 1;
	// Set the burst size.
	p_HW->DMA.BurstSizeB[ChannelIndex - 1]  = BurstSizeB;
	// Set the burst count.
	p_HW->DMA.BurstCount[ChannelIndex - 1]  = BurstCount;
	// Set the page count.
	p_HW->DMA.PageCount[ChannelIndex - 1]   = PageCount;
	// Set the pages per interrupt.
	p_HW->DMA.PagesPerMark[ChannelIndex - 1]  = PagesPerMark;
	// Alocate the requested DMA buffer.
	Status = Adapter_DMABufAllocate(p_HW);
	// If allocation was successful, load the physical address of each DMA page to the hardware.
	if (Status == 0)
	{
		// Set the DMA index register to load page offsetes for the selected channel.
		Status = Adapter_Write64(p_HW, Bar, Off - 0x8, 0x2 <<p_HW->DMA.ChannelIndex);

		// Initialize the address index to the start of the page memory.
		add_index = Off;
		// Write the DMA physical page addresses to the adapter.
		for(page_count=0; page_count < p_HW->DMA.PageCount[p_HW->DMA.ChannelIndex]; page_count++)
		{
			// Perform the write operation to the page memory.
			Status = Adapter_Write64(p_HW, Bar, add_index, p_HW->DMA.PhysDMAAdr[p_HW->DMA.ChannelIndex][page_count]);
			// Increment the page memory address.
			add_index += 0x8;
		}

		// Set the DMA index register to load descriptors for the selected channel.
		Status = Adapter_Write64(p_HW, Bar, Off - 0x8, 0x200<<p_HW->DMA.ChannelIndex);
		
		// Initialize the address index to the start of the descriptor memory.
		add_index = Off;

		// Form the descriptors and write them to the adapter.
		for(page_count=0; page_count < p_HW->DMA.PageCount[p_HW->DMA.ChannelIndex]; page_count++)
		{
			// Set the Burst Size and Burst Count fields of the descriptor.
			descriptor = (unsigned long long)(p_HW->DMA.BurstSizeB[p_HW->DMA.ChannelIndex] << 16) | (p_HW->DMA.BurstCount[p_HW->DMA.ChannelIndex]);

			//Set the Last Page descriptor bit if this is the final page of this channel.
			if (page_count == p_HW->DMA.PageCount[p_HW->DMA.ChannelIndex]-1)
				descriptor |= 0x0000000040000000ull;

			//Set the Interrupt descriptor bit if this page matches the requested pages per interrupt setting.
			if (((page_count+1) % p_HW->DMA.PagesPerMark[p_HW->DMA.ChannelIndex]) == 0)
				descriptor |= 0x0000000080000000ull;

			// Perform the write operation to the descriptor memory.
			Status = Adapter_Write64(p_HW, Bar, add_index, descriptor);
			// Increment the descriptor memory address.
			add_index += 0x8;
		}
	}

//		Status = LoadDMASettings(p_HW, Bar, Off);
	return(Status);
}

unsigned long DMAFree(s_Adapter *p_HW, unsigned long ChannelIndex)
{
	unsigned long Status = 0;

	// Set the device and channel number.
	p_HW->DMA.ChannelIndex = ChannelIndex - 1;
	// Free the requested DMA buffer.
	Adapter_DMABufFree(p_HW);
	return(Status);
}

unsigned long DMASave(s_Adapter *p_HW, unsigned long ChannelIndex, char * DataFile, size_t Bits)
{
	unsigned long Status = 0;
	FILE *fp;
	unsigned long long one = 0x1, max = 0xFFFFFFFFFFFFFFFFull;
	size_t lines_per_page;
	size_t page_count, sample_count;
	size_t line_count;
	long long sample;
	unsigned long long *virtual_addresss;

	// Set the device and channel number.
	p_HW->DMA.ChannelIndex = ChannelIndex - 1;


	// Open the specified data file.
	fp = fopen(DataFile, "w");

	// Process samples only if the file open was successful.
	if(fp != NULL)
	{
		// Compute the number of 64-bit data lines per DMA page (8 bytes per line).
		lines_per_page = p_HW->DMA.BurstSizeB[p_HW->DMA.ChannelIndex] * p_HW->DMA.BurstCount[p_HW->DMA.ChannelIndex] / 8;

		// Sequence through all of the DMA pages assigned to the channel.
		for(page_count=0; page_count < p_HW->DMA.PageCount[p_HW->DMA.ChannelIndex]; page_count++)
		{
			// Get the virtual address of the first 64-bit data line in each page.
			virtual_addresss = (unsigned long long *)p_HW->DMA.VirtDMAAdr[p_HW->DMA.ChannelIndex][page_count];

			// Sequence through every 64-bit data line in the page.
			for(line_count = 0; line_count < lines_per_page; line_count++)

				// Sequence through every sample in the 64-bit data line.
				for (sample_count = 0; sample_count < (64 / Bits); sample_count++)
					{
						// Extract the next sample value from the 64-bit data line.
						sample = (*(virtual_addresss + line_count) >> (sample_count * Bits)) & (max >> (64 - Bits));
						// Sign extend negative numbers.
						if (sample & (one << (Bits - 1)))
							sample |= (max << Bits);
						// Print the sample value to the file.
						fprintf(fp, "%lld\n", sample);
					}
		}
		fclose(fp);
		Status = 0;
	}
	else
		Status = 1;

	return(Status);
}

void HoldKeyboard()
{
	char enter;

	// Print a message to the screen.
	printf("\nPress **ENTER** from the keyboard to continue...\n\n");
	// Wait for an ENTER from the keyboard.
	while((enter = (char)getchar()) != '\n' && enter != EOF);
}

unsigned long HoldTimer(unsigned long Time)
{
	unsigned long Status = 0;

	// Hold for the specified time.
	Status = Adapter_uSleep(Time);
	return(Status);
}

unsigned long long ModifyOR(unsigned long long Value, unsigned long long Modifier)
{
	unsigned long long Result = 0;

	Result = Value | Modifier;
	return(Result);
}

unsigned long long ModifyAND(unsigned long long Value, unsigned long long Modifier)
{
	unsigned long long Result = 0;

	Result = Value & Modifier;
	return(Result);
}

unsigned long Query(s_Adapter *p_HW, unsigned long Bar, unsigned long Off, char *Operator, unsigned long long Operand, unsigned long Attempts, unsigned long Delay, unsigned long long *Value)
{
	unsigned long Status = 0;
	unsigned long cnt = 0;
	int logical;

	// Read the value in the register.
	Status = Adapter_Read64(p_HW, Bar, Off, Value);
	// Perform the relational operation.
	logical = Conditional(Operator, Operand, *Value);
	// Make additional attempts until the relational is satisfied or the maximum number of attempts is reached.
	while(logical && (cnt != Attempts))
	{
		// Pause for the specified time.
		Adapter_uSleep(Delay);
		// Read the value in the register.
		Status = Adapter_Read64(p_HW, Bar, Off, Value);
		// Increment the count of attempts.
		cnt++;
		// Perform the relational operation.
		logical = Conditional(Operator, Operand, *Value);
	}
	return(Status);
}

unsigned long Read(s_Adapter *p_HW, unsigned long Bar, unsigned long Off, unsigned long long *Value)
{
	unsigned long Status = 0;

	// Read the requested offset and store the result in Value.
	Status = Adapter_Read64(p_HW, Bar, Off, Value);
	return(Status);
}

unsigned long Serial(s_Adapter *p_HW, unsigned long Bar, unsigned long Off, unsigned long long Configure, unsigned long long Payload, unsigned long long *p_ReadValue)
{
	unsigned long Status = 0;
	unsigned long long config_payload = 0;
	unsigned long long max = 0xFFFFFFFFFFFFFFFFull;

	// Create a single 64-bit word by concatonating the 8-bit configuration and 56-bit payload constants.
	config_payload = (Configure << 56) | Payload;

	// Load the payload and configuration bits.
	Status += Write(p_HW, Bar, Off, config_payload);

	// Initiate a SPI transaction by writing any value to the SPI start address.
	Status += Write(p_HW, Bar, Off + 0x8, 0x0ull);

	// Wait for the serial transaction to complete by monitoring the busy bit.
	do
		Status += Read(p_HW, Bar, Off + 0x8, p_ReadValue);
	while(*p_ReadValue);
	
	// Read the payload register to retrieve the results of a serial read operation.
	Status += Read(p_HW, Bar, Off, p_ReadValue);

	// Mask the bits to match the payload size.
	switch (Configure & 0x07ull)
	{
	case 0: {*p_ReadValue &= (max >> 56); break;}
	case 1: {*p_ReadValue &= (max >> 48); break;}
	case 2: {*p_ReadValue &= (max >> 40); break;}
	case 3: {*p_ReadValue &= (max >> 32); break;}
	case 4: {*p_ReadValue &= (max >> 24); break;}
	case 5: {*p_ReadValue &= (max >> 16); break;}
	default: {Status++; break;}
	}

	// Return an error code if the program operation did not complete.
	if (Status)
		return(1);
	else
		return(0);
}

unsigned long Write(s_Adapter *p_HW, unsigned long Bar, unsigned long Off, unsigned long long Value)
{
	unsigned long Status = 0;

	// Read the requested offset and store the result in Value.
	Status = Adapter_Write64(p_HW, Bar, Off, Value);
	return(Status);
}

unsigned long Conditional(char *Operator, unsigned long long Operand, unsigned long long Value)
{
	int decision = -1;

	switch (Operator[0])
	{
	case '=':
		{
			if (Value == Operand)
				decision = 0;
			break;
		}
	case '!':
		{
			if (Value != Operand)
				decision = 0;
			break;
		}
	case '<':
		{
			if (strlen(Operator) == 1)
			{
				if (Value < Operand)
					decision = 0;
			}
			else
				if (Value <= Operand)
					decision = 0;
			break;
		}
	case '>':
		{
			if (strlen(Operator) == 1)
			{
				if (Value > Operand)
					decision = 0;
			}
			else
				if (Value >= Operand)
					decision = 0;
			break;
		}
	case '+':
		{
			if (strlen(Operator) == 1)
			{
				if ((Value & Operand) > 0x0)
					decision = 0;
			}
			else
				if ((~Value & Operand) == 0x0)
					decision = 0;
			break;
		}
	case '-':
		{
			if (strlen(Operator) == 1)
			{
				if ((~Value & ~Operand) > 0x0)
					decision = 0;
			}
			else
				if ((Value & ~Operand) == 0x0)
					decision = 0;
			break;
		}
	}

	return decision;
}

